//------------------------------------------------------------------------------------//
//includes
#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

extern "C" BOOL WINAPI _DllMainCRTStartup(HMODULE,DWORD,LPVOID) { return 1; }

#define _WIN32_DCOM
#include <objbase.h>
#include <dvoice.h>
#include <dplay8.h>

//------------------------------------------------------------------------------------//
//macros and type definitions

typedef struct
{
	DWORD  mVersion;
	HWND   mHwnd;
	BOOL   mKeep;
} LOADINFO;

#define MAX_PLAYERS 8
#define NOTIFY_PERIOD 50

#define CMD_MUTE		561
#define CMD_KICK		593
#define CMD_DATA		557
#define CMD_HOST		622
#define CMD_TALK		663
#define CMD_SEND		692
#define CMD_CLOSE		1135
#define CMD_CONNECT		4502
#define CMD_HANDLER		4840
#define CMD_TARGETS		5175
#define CMD_VERSION		5260
#define CMD_PASSWORD	11156

#define ret(x) { lstrcpy(data,x); return 3; }

#define CREATE_STRUCT(type,name) \
	type name; \
	ZeroMemory(&name,sizeof(name)); \
name.dwSize = sizeof(name);

//------------------------------------------------------------------------------------//
//synchronization objects
HANDLE					filemap_mutex;		//for SendMessage

//------------------------------------------------------------------------------------//
//file mapping

HWND					mWnd;					//mIRC window
HANDLE					hFileMap;				//mapped file handle
char					*mData;					//mapped file contents

//------------------------------------------------------------------------------------//
//session info

wchar_t					u_password[1024];		//session password
wchar_t					u_peer_info[1024]		= L"nPLAYn 1.0 user";

char					alias[1024];			//event handler
char					context[1024];			//context string

//------------------------------------------------------------------------------------//
// DirectPlay data
IDirectPlayVoiceServer	*vserver;				//the one who controls voice session
IDirectPlayVoiceClient	*vclient;				//the one who speaks
IDirectPlay8Peer		*peer;					//the one who resets connections ;)
int	avg_input, avg_output, tact;				//volume control

DPN_APPLICATION_DESC app_desc =
{
	sizeof(DPN_APPLICATION_DESC),
	DPNSESSION_MIGRATE_HOST | DPNSESSION_REQUIREPASSWORD,
	{ 0 },
	{ 0xa75ef3a0, 0xfc87, 0x11d5, { 0xac, 0xc8, 0x44, 0x45, 0x53, 0x54, 0x0, 0x0 } },
	MAX_PLAYERS,0, 0,u_password
};

//------------------------------------------------------------------------------------//
//other players info
char		peer_info[MAX_PLAYERS][1024];
DPNHANDLE	peer_handles[MAX_PLAYERS];

//------------------------------------------------------------------------------------//
//minimal CRT

DWORD number(char**pp)
{
	int result = 0, sign = 1;
	if (**pp == '-') { sign = -1; (*pp)++; }
	while (**pp >= '0' && **pp <= '9') { result *= 10; result += (*(*pp)++ - '0'); }
	while(**pp && **pp != ' ') (*pp)++;
	while(**pp == ' ') (*pp)++;
	return result * sign;
}

//------------------------------------------------------------------------------------//
//calls event handler
void mirc(const char *msg,char *outbuf = 0)
{
	if (*alias)
	{
		if (WaitForSingleObject(filemap_mutex,10000) != WAIT_TIMEOUT)
		{
			wsprintf(mData,"$isalias(%s)",alias);
			SendMessage(mWnd,WM_USER + 201,0,0);
			if (mData[1] == 't')
			{
				if (outbuf)
				{
					wsprintf(mData,"$%s(%s,peer,%s)",alias,context,msg);
					SendMessage(mWnd,WM_USER + 201,0,0);
					lstrcpy(outbuf,mData);
				}
				else
				{
					wsprintf(mData,"/%s %s peer %s",alias,context,msg);
					SendMessage(mWnd,WM_USER + 200,1,0);
				}
			}
			else *alias = 0;
			ReleaseMutex(filemap_mutex);
		}
	}
}
//------------------------------------------------------------------------------------//
HRESULT set_peer_info(DWORD dwFlags)
{
	CREATE_STRUCT(DPN_PLAYER_INFO,pi);
	pi.dwInfoFlags = DPNINFO_NAME;
	pi.pwszName = u_peer_info;
	DPNHANDLE async_handle;
	return peer->SetPeerInfo(&pi,0,&async_handle,dwFlags);
}
//------------------------------------------------------------------------------------//
void get_peer_info(DPNID player,char *info)
{
	union { DPN_PLAYER_INFO pi; char buffer[2048]; } peer_info;
	DWORD size = 2048;
	peer_info.pi.dwSize = sizeof(DPN_PLAYER_INFO);
	
	//GetPeerInfo might return DPNERR_CONNECTING when connecting, so just keep calling it if it does
	HRESULT hr;
	do { hr = peer->GetPeerInfo(player,&(peer_info.pi),&size,0); } while(hr == DPNERR_CONNECTING);
	WideCharToMultiByte(CP_ACP,0,FAILED(hr) ? L"nPLAYn user" : peer_info.pi.pwszName,-1,info,1024,0,0);
}
//------------------------------------------------------------------------------------//
HRESULT WINAPI voice_message_handler(void*,DWORD message,void* buf)
{
	char msg[1024];
	switch(message)
	{
	case DVMSGID_CREATEVOICEPLAYER:
		{
			PDVMSG_CREATEVOICEPLAYER p = PDVMSG_CREATEVOICEPLAYER(buf);
			for (int i = 0; i < MAX_PLAYERS; i++)
			{
				if (peer_handles[i] == p->dvidPlayer)
				{
					p->pvPlayerContext = (void*)(i+1);
					wsprintf(msg,"speaking %d %s",i+1,peer_info[i]);
					mirc(msg);
					break;
				}
			}
		}
		break;
	case DVMSGID_DELETEVOICEPLAYER:
		{
			PDVMSG_DELETEVOICEPLAYER p = PDVMSG_DELETEVOICEPLAYER(buf);
			wsprintf(msg,"silent %d %s",p->pvPlayerContext,peer_info[int(p->pvPlayerContext)-1]);
			mirc(msg);
		}
		break;
	case DVMSGID_INPUTLEVEL:
		avg_input += PDVMSG_INPUTLEVEL(buf)->dwPeakLevel;
		break;
	case DVMSGID_OUTPUTLEVEL:
		avg_output += PDVMSG_OUTPUTLEVEL(buf)->dwPeakLevel;
		if (++tact == NOTIFY_PERIOD / 10)
		{
			wsprintf(msg,"volume %d %d",int(avg_input / 5),int(avg_output / 5));
			mirc(msg);
			avg_output = avg_input = tact = 0;
		}
		break;
	case DVMSGID_CONNECTRESULT:
		mirc(FAILED(LPDVMSG_CONNECTRESULT(buf)->hrResult) ? "cannot_talk" : "talking");
		break;
	case DVMSGID_DISCONNECTRESULT:
		mirc(FAILED(LPDVMSG_DISCONNECTRESULT(buf)->hrResult) ? "cannot_mute" : "muted");
		break;
	case DVMSGID_HOSTMIGRATED:
		if (PDVMSG_HOSTMIGRATED(buf)->pdvServerInterface)
		{
			vserver->Release();
			vserver = PDVMSG_HOSTMIGRATED(buf)->pdvServerInterface;
			vserver->AddRef();
		}
		break;
	case DVMSGID_RECORDSTART:
		mirc("voice_on");
		break;
	case DVMSGID_RECORDSTOP:
		mirc("voice_off");
		break;
	case DVMSGID_GAINFOCUS:
		mirc("focus_gained");
		break;
	case DVMSGID_LOSTFOCUS:
		mirc("focus_lost");
		break;
	}
	return S_OK;
}

//------------------------------------------------------------------------------------//
HRESULT WINAPI peer_message_handler(void*,DWORD message,void* buf)
{
	char msg[1024];
	switch(message)
	{
	case DPN_MSGID_CONNECT_COMPLETE:
		mirc(FAILED(PDPNMSG_CONNECT_COMPLETE(buf)->hResultCode) ? "cannot_connect" : "connected");
		break;
	case DPN_MSGID_CREATE_PLAYER:
		{
			PDPNMSG_CREATE_PLAYER p = PDPNMSG_CREATE_PLAYER(buf);
			for (int i = 0; i < MAX_PLAYERS; i++)
			{
				if (!peer_handles[i])
				{
					p->pvPlayerContext = (void*)(i+1);
					peer_handles[i] = p->dpnidPlayer;
					get_peer_info(p->dpnidPlayer,peer_info[i]);
					wsprintf(msg,"joined %d %s",i+1,peer_info[i]);
					mirc(msg);
					break;
				}
			}
		}
		break;
	case DPN_MSGID_DESTROY_PLAYER:
		{
			PDPNMSG_DESTROY_PLAYER p = PDPNMSG_DESTROY_PLAYER(buf);
			wsprintf(msg,"parted %d %s",p->pvPlayerContext,peer_info[int(p->pvPlayerContext)-1]);
			peer_handles[int(p->pvPlayerContext)-1] = 0;
			mirc(msg);
		}
		break;
	case DPN_MSGID_INDICATE_CONNECT:
		{
			PDPNMSG_INDICATE_CONNECT p = PDPNMSG_INDICATE_CONNECT(buf);
			wchar_t unicode_hostname[1024];
			DWORD size = 1024, type;
			p->pAddressPlayer->AddRef();
			p->pAddressPlayer->GetComponentByName(DPNA_KEY_HOSTNAME,unicode_hostname,&size,&type);
			p->pAddressPlayer->Release();
			wsprintf(msg,"confirm %ws",unicode_hostname);
			mirc(msg,(char*)unicode_hostname);
			if (*((char*)unicode_hostname) == 'E') return E_FAIL;
		}
		break;
	case DPN_MSGID_PEER_INFO:
		{
			PDPNMSG_PEER_INFO p = PDPNMSG_PEER_INFO(buf);
			get_peer_info(p->dpnidPeer,peer_info[int(p->pvPlayerContext)-1]);
			wsprintf(msg,"new_data %d %s",p->pvPlayerContext,peer_info[int(p->pvPlayerContext)-1],1024);
			mirc(msg);
		}
		break;
	case DPN_MSGID_RECEIVE:
		{
			PDPNMSG_RECEIVE p = PDPNMSG_RECEIVE(buf);
			wsprintf(msg,"received %d %s",p->pvPlayerContext,p->pReceiveData);
			mirc(msg);
		}
		break;
	case DPN_MSGID_SEND_COMPLETE:
		{
			PDPNMSG_SEND_COMPLETE p = PDPNMSG_SEND_COMPLETE(buf);
			if (FAILED(p->hResultCode))
				wsprintf(msg,"cannot_send %d %d",p->pvUserContext,p->dwSendTime); 
			else
				wsprintf(msg,"sent %d %d",p->pvUserContext,p->dwSendTime); 
			mirc(msg);
		}
		break;
	case DPN_MSGID_TERMINATE_SESSION:
		{
			PDPNMSG_TERMINATE_SESSION p = PDPNMSG_TERMINATE_SESSION(buf);
			if (p->hResultCode == DPNERR_HOSTTERMINATEDSESSION)
			{
				wsprintf(msg,"kicked %s",p->pvTerminateData);
				mirc(msg);
			}
			else mirc("E_HOST_LEFT");
		}
		break;
	case DPN_MSGID_HOST_MIGRATE:
		{
			PDPNMSG_HOST_MIGRATE p = PDPNMSG_HOST_MIGRATE(buf);
			wsprintf(msg,"new_host %d %s",p->pvPlayerContext,peer_info[int(p->pvPlayerContext)-1]);
			mirc(msg);
		}
	}
	return S_OK;
}
//------------------------------------------------------------------------------------//
extern "C" void __stdcall LoadDll(LOADINFO *p_li)
{
	mWnd = p_li->mHwnd;
	
	//create file mapping
	filemap_mutex = CreateMutex(0,FALSE,"mIRC_SendMessage");
	hFileMap = CreateFileMapping(INVALID_HANDLE_VALUE,0,PAGE_READWRITE,0,0x4000,"mIRC");
	mData = (LPSTR)MapViewOfFile(hFileMap,FILE_MAP_ALL_ACCESS,0,0,0);
	
	//initialize DirectPlay objects
	if(SUCCEEDED(CoCreateInstance(CLSID_DirectPlay8Peer,0,CLSCTX_INPROC_SERVER,IID_IDirectPlay8Peer,(void**)&peer)))
	{
		peer->Initialize(0,peer_message_handler,0);
		CoCreateInstance(CLSID_DirectPlayVoiceServer,0,CLSCTX_INPROC_SERVER,IID_IDirectPlayVoiceServer,(void**)&vserver);
		CoCreateInstance(CLSID_DirectPlayVoiceClient,0,CLSCTX_INPROC_SERVER,IID_IDirectPlayVoiceClient,(void**)&vclient);
		vserver->Initialize(peer,0,0,0,0);
		vclient->Initialize(peer,voice_message_handler,0,0,0);
	}
}

//------------------------------------------------------------------------------------//
extern "C" int __stdcall UnloadDll(int timeout)
{
	if (!timeout)
	{
		*alias = 0;	//no more messages :(
		
		//shutdown peer
		if (vclient) { vclient->Disconnect(0); vclient->Release(); }
		if (vserver) { vserver->StopSession(0); vserver->Release(); }
		if (peer) { peer->Close(0); peer->Release(); }
		
		//delete file mapping
		UnmapViewOfFile(mData);
		CloseHandle(hFileMap);
		CloseHandle(filemap_mutex);
	}
	return 0;
}

//------------------------------------------------------------------------------------//
extern "C" int __stdcall call(HWND mWnd,HWND,char*data,char*,BOOL show,BOOL)
{
	//hash command token
	int hash = 0;
	for(char *p = data; *p && *p != ' '; p++) hash = (hash << 1) ^ (*p | 32);
	if (hash == CMD_VERSION) ret("nPLAYn 1.0 by Necroman - #mIRC @ Undernet");
	if (!vclient) ret("E_INSTALL_DX8");
	
	while(*p == ' ') p++;
	switch(hash)
	{
	case CMD_HOST:
		{
			//set peer info
			set_peer_info(DPNSETPEERINFO_SYNC);
			
			//read codec and port
			DWORD codec = number(&p), port = number(&p);
			if (codec < 1 || codec > 7) ret("E_WRONG_CODEC");
			
			//create device address
			IDirectPlay8Address *device;
			CoCreateInstance(CLSID_DirectPlay8Address,0,CLSCTX_INPROC_SERVER,IID_IDirectPlay8Address,(void**)&device);
			device->SetSP(&CLSID_DP8SP_TCPIP);
			if (port)
			{
				device->AddComponent(DPNA_KEY_PORT,&port,sizeof(port),DPNA_DATATYPE_DWORD);
				app_desc.dwFlags |= DPNSESSION_NODPNSVR;
			} else app_desc.dwFlags &= ~DPNSESSION_NODPNSVR;
			
			//call Host
			HRESULT hr;
			hr = peer->Host(&app_desc,&device,1,0,0,0,DPNCONNECT_OKTOQUERYFORADDRESSING);
			device->Release();
			if (FAILED(hr)) ret("E_CANNOT_TALK");
			
			//start voice session
			CREATE_STRUCT(DVSESSIONDESC,sd);
			sd.dwSessionType = DVSESSIONTYPE_PEER; 
			GUID c_types[7];
			c_types[0] = DPVCTGUID_VR12;
			c_types[1] = DPVCTGUID_SC03;
			c_types[2] = DPVCTGUID_SC06;
			c_types[3] = DPVCTGUID_TRUESPEECH;
			c_types[4] = DPVCTGUID_GSM;
			c_types[5] = DPVCTGUID_ADPCM;
			c_types[6] = DPVCTGUID_NONE;
			sd.guidCT = c_types[codec-1];
			ret(FAILED(vserver->StartSession(&sd,0)) ? "E_HOST_NO_SOUND" : "S_HOSTING");
		}
		
	case CMD_CONNECT:
		{
			//set peer info
			app_desc.dwFlags &= ~DPNSESSION_NODPNSVR;
			set_peer_info(DPNSETPEERINFO_SYNC);
			
			//create device address
			IDirectPlay8Address *device;
			CoCreateInstance(CLSID_DirectPlay8Address,0,CLSCTX_INPROC_SERVER,IID_IDirectPlay8Address,(void**)&device);
			device->SetSP(&CLSID_DP8SP_TCPIP);
			
			//create host address
			IDirectPlay8Address *host;
			CoCreateInstance(CLSID_DirectPlay8Address,0,CLSCTX_INPROC_SERVER,IID_IDirectPlay8Address,(void**)&host);
			host->SetSP(&CLSID_DP8SP_TCPIP);
			
			//add hostname component
			char hostname[1024], *a = hostname; *a = 0;
			while(*p && *p != ' ') *a++ = *p++; *a++ = 0;
			wchar_t u_hostname[1024];
			MultiByteToWideChar(CP_ACP,0,hostname,-1,u_hostname,1024);
			if (*hostname)
				host->AddComponent(DPNA_KEY_HOSTNAME,u_hostname,(a - hostname)<<1,DPNA_DATATYPE_STRING);
			
			//add host portnum
			while(*p == ' ') p++;
			DWORD remote_port = number(&p), local_port = number(&p);
			if (remote_port)
				host->AddComponent(DPNA_KEY_PORT,&remote_port,sizeof(DWORD),DPNA_DATATYPE_DWORD);
			if (local_port)
				device->AddComponent(DPNA_KEY_PORT,&local_port,sizeof(DWORD),DPNA_DATATYPE_DWORD);
			
			//call Connect
			DPNHANDLE async_handle;				
			HRESULT hr = peer->Connect(&app_desc,host,device,0,0,0,0,0,0,&async_handle,DPNCONNECT_OKTOQUERYFORADDRESSING);
			device->Release(); host->Release();
			ret(FAILED(hr) ? "E_CANNOT_CONNECT" : "S_CONNECTING");		
		}
		
	case CMD_TALK:
		{
			HWND window = (HWND)number(&p); if (!IsWindow(window)) window = mWnd;
			CREATE_STRUCT(DVSOUNDDEVICECONFIG,sdc);
			sdc.dwFlags = DVSOUNDCONFIG_AUTOSELECT | DVSOUNDCONFIG_SETCONVERSIONQUALITY;
			sdc.guidPlaybackDevice        = DSDEVID_DefaultVoicePlayback; 
			sdc.guidCaptureDevice         = DSDEVID_DefaultVoiceCapture; 
			sdc.hwndAppWindow             = window;
			
			CREATE_STRUCT(DVCLIENTCONFIG,cc);
			cc.dwFlags                = DVCLIENTCONFIG_AUTOVOICEACTIVATED | DVCLIENTCONFIG_AUTORECORDVOLUME;
			cc.lPlaybackVolume        = DVPLAYBACKVOLUME_DEFAULT;
			cc.dwThreshold	          = DVTHRESHOLD_UNUSED;
			cc.lRecordVolume          = DVRECORDVOLUME_LAST;
			cc.dwNotifyPeriod         = NOTIFY_PERIOD;
			
			//connect
			DVID all = DVID_ALLPLAYERS;
			HRESULT hr = vclient->Connect(&sdc,&cc,DVFLAGS_SYNC);
			if (hr == DVERR_RUNSETUP)
			{
				char reply[1024];
				mirc("run_setup",reply);
				if ('E' == *reply) ret("E_RUN_SETUP");
				IDirectPlayVoiceTest *test;
				CoCreateInstance(CLSID_DirectPlayVoiceTest,0,CLSCTX_INPROC_SERVER,IID_IDirectPlayVoiceTest,(void**)&test);
				HRESULT hr = test->CheckAudioSetup(0,0,mWnd,0);
				test->Release();
				if (hr == DV_FULLDUPLEX) mirc("full_duplex");
				if (hr == DV_HALFDUPLEX) mirc("half_duplex");
				if (FAILED(hr)) ret("E_SETUP_FAILED");
				hr = vclient->Connect(&sdc,&cc,DVFLAGS_SYNC);
			}
			if (FAILED(hr)) ret("E_CANNOT_TALK");
			vclient->SetTransmitTargets(&all,1,0);
			ret("S_TALKING");
		}

	case CMD_CLOSE:
		vclient->Disconnect(0);
		vserver->StopSession(0);
		peer->Close(0);
		peer->Initialize(0,peer_message_handler,0);
		ret("S_SESSION_CLOSED");
	
	case CMD_TARGETS:
		{
			int i = 0;
			DVID targets[MAX_PLAYERS];
			while (*p && i <= MAX_PLAYERS)
			{
				DWORD player = number(&p);
				if (!player) { *targets = 0; i = 1; break; }
				if (player > MAX_PLAYERS || !peer_handles[player-1]) ret("E_WRONG_TARGETS");
				targets[i++] = (DVID)peer_handles[player-1];
			}
			ret(FAILED(vclient->SetTransmitTargets(i ? targets : 0,i,0)) ? "E_CANNOT_SET_TARGETS" : "S_TARGETS_SET");
		}
	case CMD_MUTE:
		vclient->Disconnect(0);
		ret("S_VOICE_OFF");
		
	case CMD_SEND:
		{
			DWORD player_num = number(&p), context = number(&p), timeout = number(&p), options = number(&p);
			if (player_num > MAX_PLAYERS) ret("E_WRONG_TARGET");
			BUFFERDESC bd = { lstrlen(p)+1, (LPBYTE)p };
			DPNHANDLE async_handle;
			ret(FAILED(peer->SendTo(player_num ? peer_handles[player_num-1] : 0,&bd,1,timeout,(void*)context,&async_handle,options)) ? "E_CANNOT_SEND" : "S_SENDING");
		}
		
	case CMD_HANDLER:
		lstrcpy(alias,p);
		while(*p && *p != ' ') p++; while(*p == ' ') p++;
		lstrcpy(context,*p ? p : "nPLAYn");
		ret("S_HANDLER_SET");
		
	case CMD_DATA:
		{
			MultiByteToWideChar(CP_ACP,0,p,-1,u_peer_info,1024);
			ret(FAILED(set_peer_info(0)) ? "E_CANNOT_SET_DATA" : "S_DATA_SET");
		}
		
	case CMD_PASSWORD:
		MultiByteToWideChar(CP_ACP,0,p,-1,u_password,1024);
		ret("S_PASSWORD_SET");
		
	case CMD_KICK:
		{
			DWORD player = number(&p);
			if (!player || player > MAX_PLAYERS || !peer_handles[player-1]) ret("E_WRONG_VICTIM");
			HRESULT hr = peer->DestroyPeer(peer_handles[player-1],p,lstrlen(p)+1,0);
			if (hr == DPNERR_NOTHOST) ret("E_NOT_HOST");
			ret(FAILED(hr) ? "E_CANNOT_KICK" : "S_KICKING");
		}		
	}
	ret("E_NOT_IMPLEMENTED");
}